import cctbx.miller

def extract_structure_factors(mtz_object, ampl_label, phas_label):
    # Get the crystal symmetry from the amplitudes' crystal
    try:
        ampl_col = mtz_object.get_column(ampl_label)
    except:
        raise Exception('Amplitude column not found: {}. Have you specified the right columns?'.format(ampl_label))
    crystal_symmetry = ampl_col.mtz_crystal().crystal_symmetry()
    # Create miller set
    mill_idx = mtz_object.extract_miller_indices()
    mill_set = cctbx.miller.set(crystal_symmetry=crystal_symmetry, indices=mill_idx)
    # Extract amplitudes and phases
    try:
        ampl_com = mtz_object.extract_complex(column_label_ampl=ampl_label, column_label_phi=phas_label)
    except:
        raise Exception('Could not extract structure factors - Amplitudes:{}, Phases:{}. Have you specified the right columns?'.format(ampl_label, phas_label))
    # Convert to miller array
    mill_sfs = mill_set.array(ampl_com.data)
    mill_sfs.set_observation_type_xray_amplitude()
    # Check it's complex
    assert mill_sfs.is_complex_array(), 'STRUCTURE FACTORS SHOULD BE COMPLEX?!'
    # Make non-anomalous
    mill_sfs = mill_sfs.as_non_anomalous_array()
    return mill_sfs



from __future__ import print_function

import os, sys, glob, time, re

#################################
try:
    import matplotlib
    matplotlib.use('Agg')
    matplotlib.interactive(0)
    from matplotlib import pyplot
    pyplot.style.use('ggplot')
except:
    pass
#################################

import numpy, pandas

import iotbx.pdb

from scitbx.array_family import flex

from bamboo.common import Meta, Info
from bamboo.plot import simple_histogram, simple_bar, simple_scatter

from bamboo.maths.angles import mean_std_angles

from giant.structure import make_label
from giant.structure.dihedrals import get_all_phi_psi_for_hierarchy
from giant.structure.b_factors import BfactorStatistics, normalise_b_factors_to_z_scores
from giant.structure.iterators import residues_via_conformers, conformers_via_residue_groups

import giant.structure.select as s_select

class StructureCollection(object):
    """Class to hold information about a set of related structures"""

    def __init__(self, hierarchy_inputs=None, hierarchies=None, inputs=None, labels=None):

        # Object to stores all of the structure objects
        self.structures = Info(['hierarchies','inputs','labels'])
        # Unpack and store structure objects
        if hierarchy_inputs:
            self.structures.hierarchies = [i.hierarchy for i in hierarchy_inputs]
            self.structures.inputs      = [i.input     for i in hierarchy_inputs]
        else:
            self.structures.hierarchies = hierarchies
            self.structures.inputs      = inputs
        # Unpack and store labels
        if not labels: labels = range(len(self.structures.hierarchies))
        self.structures.labels = labels

        #Initialise output tables
        self.initialise_tables()

    @classmethod
    def from_files(cls, filenames, labels=None):
        """Initialise the object from files"""
        if not labels: labels=filenames
        return cls(hierarchy_inputs=[iotbx.pdb.hierarchy.input(f) for f in filenames], labels=labels)

    ######################################################################################

    def initialise_tables(self):
        """Populate the index of the tables"""
        # Object to contain all of the output data tables
        self.tables = Info(['structures','residues','atoms'])
        self._initialise_structure_table()
        self._initialise_residue_table()

    def _initialise_structure_table(self):
        """Initialise the tables.structures object with structure labels"""
        self.tables.structures = pandas.DataFrame( data = None,
                                                   index   = pandas.Index(data=self.structures.labels, name='structure label'),
                                                   columns = []      )

    def _initialise_residue_table(self):
        """Initialise the tables.residues object with residue labels"""
        residue_labels = [make_label(c) for c in conformers_via_residue_groups(self.structures.hierarchies[0])]
        # Multiple observations of the same residue
        self.tables.residue_observations = pandas.Panel( data       = None,
                                                         items      = pandas.Index(data=residue_labels, name=['chain','residue','altloc'], tupleize_cols=True),
                                                         major_axis = pandas.Index(data=self.structures.labels, name='structure label'),
                                                         minor_axis = pandas.Index(data=[], name='variables')     )
        # Condensed statistics for each residue
        self.tables.residue_statistics = pandas.DataFrame( data = None,
                                                           index = pandas.Index(data=residue_labels, name=['chain','residue','altloc'], tupleize_cols=True),
                                                           columns = []      )

    ######################################################################################

    def load_all(self):
        self.load_all_structure_data()
        self.load_all_residue_data()
    def load_all_structure_data(self):
        self.extract_structure_info()
        self.calculate_structure_mean_b_factors()
    def load_all_residue_data(self):
        self.extract_residue_info()
        self.calculate_residue_mean_b_factors()
        self.calculate_residue_mean_normalised_b_factors()
        self.calculate_residue_phi_psi_angles()
        self.reduce_residue_bfactor_observations()
        self.reduce_residue_phi_psi_observations()

    def write_all(self, out_dir):
        self.write_tables(out_dir=out_dir)
        self.write_residue_statistics_as_structures(out_dir=out_dir)
        self.write_normalised_structures(out_dir=out_dir)
        self.write_graphs(out_dir=out_dir)

    ######################################################################################

    def write_tables(self, out_dir):
        """Dump all data as csv files in out_dir"""
        if not os.path.exists(out_dir): os.mkdir(out_dir)
        print('------------------------------------>')
        # Write csv for all global structure variables
        filename = os.path.join(out_dir, 'all_structures.csv')
        print('Writing {}'.format(filename))
        self.tables.structures.to_csv(filename)
        # Write csv for all global residue variables
        filename = os.path.join(out_dir, 'all_residues.csv')
        print('Writing {}'.format(filename))
        self.tables.residue_statistics.to_csv(filename)
        # Write csv for each variable for residues
        for variable in self.tables.residue_observations.minor_axis:
            filename = os.path.join(out_dir, 'all_'+variable+'.csv')
            print('Writing {}'.format(filename))
            self.tables.residue_observations.xs(variable, axis=2).T.to_csv(filename)

    def write_residue_statistics_as_structures(self, out_dir):
        """Write out structures with different residue statistics as the b-factors"""
        if not os.path.exists(out_dir): os.mkdir(out_dir)
        print('------------------------------------>')
        blank_h = self.structures.hierarchies[0].deep_copy()
        blank_h.atoms().set_b(flex.double([0]*blank_h.atoms_size()))
        cache = blank_h.atom_selection_cache()
        for variable in self.tables.residue_statistics.columns:
            out_pdb = os.path.join(out_dir, 'variation-{}.pdb'.format(variable))
            print('Writing {} as b-factors: {}'.format(variable, out_pdb))
            # Create new copy of the structure
            var_h = blank_h.deep_copy()

            for res_lab in self.tables.residue_statistics.index:
                # Get the value of the statistic for the residue
                statistic = self.tables.residue_statistics.get_value(res_lab, variable)
                if numpy.isnan(statistic): continue
                # Get the residue atoms for the residue
                selection = 'chain "{0}" and resid "{1}" and (altloc " " or altloc "{2}")'.format(*res_lab)
                residue = var_h.select(cache.selection(selection))
                # Set the b-factors of the residue to the statistic - WILL OVERRIDE OTHER CONFORMERS
                residue.atoms().set_b(flex.double([statistic]*residue.atoms_size()))
                # Write the output structure
                var_h.write_pdb_file(out_pdb)

    def write_normalised_structures(self, out_dir):
        """Write out the structures with normalised B-factors"""
        if not os.path.exists(out_dir): os.mkdir(out_dir)
        print('------------------------------------>')
        for lab_h, pdb_h in zip(self.structures.labels, self.structures.hierarchies):
            out_pdb = os.path.join(out_dir, '{}-normalised-b.pdb'.format(lab_h))
            print('Writing Normalised B-Factors: {}'.format(lab_h))
            pdb_h_z = normalise_b_factors_to_z_scores(pdb_hierarchy=pdb_h, method='protein')
            pdb_h_z.write_pdb_file(file_name=out_pdb)

    def write_graphs(self, out_dir):
        """Dump histograms in out_dir"""
        if not os.path.exists(out_dir): os.mkdir(out_dir)
        print('------------------------------------>')
        # Iterate through variables and plot for all of the structures
        for variable in self.tables.structures.columns:
            # Extract observations for this variable
            col_data = self.tables.structures[variable]
            filename = os.path.join(out_dir,'structure-{}.png'.format(variable))
            if col_data.any() and (col_data.nunique()>1):
                print('Writing {}'.format(filename))
                simple_histogram( filename = filename,
                                  data  = col_data.dropna(),
                                  title = 'histogram for {!s}'.format(variable),
                                  x_lab = variable  )
        print('------------------------------------>')
        # Iterate through variables and plot for all of the residues
        for variable in self.tables.residue_statistics.columns:
            col_data = self.tables.residue_statistics[variable]
            filename = os.path.join(out_dir,'residue-{}.png'.format(variable))
            if col_data.any() and (col_data.nunique()>1):
                col_data = col_data.dropna()
                print('Writing {}'.format(filename))
                simple_bar( filename = filename,
                            y_vals   = col_data.values,
                            x_labels = col_data.index,
                            title = 'bar for {!s}'.format(variable),
                            x_lab = 'residue',
                            y_lab = variable  )
        print('------------------------------------>')
        # Plot residue variables against each other for all of the residues
        for var_1 in self.tables.residue_statistics.columns:
            col_data_1 = self.tables.residue_statistics[var_1]
            if not col_data_1.any(): continue
            for var_2 in self.tables.residue_statistics.columns:
                if var_1 == var_2: break
                col_data_2 = self.tables.residue_statistics[var_2]
                if not col_data_2.any(): continue
                filename = os.path.join(out_dir,'residue-{}-{}.png'.format(var_1, var_2))
                print('Writing {}'.format(filename))
                simple_scatter( filename = filename,
                                x_vals = col_data_1.values,
                                y_vals = col_data_2.values,
                                title = 'bar for {!s} against {!s}'.format(var_1, var_2),
                                x_lab = var_1,
                                y_lab = var_2  )
        print('------------------------------------>')
        # Iterate through variables for observations of the residues
        for variable in self.tables.residue_observations.minor_axis:
            # Extract observations for this variable
            var_data = self.tables.residue_observations.xs(variable, 2)
            # Iterate through residues and plot for variable values
            for res_lab in self.tables.residue_observations.items:
                filename = os.path.join(out_dir,'{}-{}.png'.format(variable, '_'.join(res_lab)))
                col_data = var_data[res_lab]
                if col_data.any() and (col_data.nunique()>1):
                    print('Writing {}'.format(filename))
                    simple_histogram( filename = filename,
                                      data  = col_data.dropna(),
                                      title = '{} histogram for {!s}'.format(variable, res_lab),
                                      x_lab = variable  )

    ######################################################################################

    def extract_structure_info(self):
        """Extract information from the input objects, if given"""
        if not self.structures.inputs: raise Exception('No input objects given')
        # ----------------------------------------------------->
        self.tables.structures['r-work'] = numpy.nan
        self.tables.structures['r-free'] = numpy.nan
        self.tables.structures['res-high'] = numpy.nan
        self.tables.structures['res-low']  = numpy.nan
        # ----------------------------------------------------->
        print('------------------------------------>')
        for lab_h, pdb_i in zip(self.structures.labels, self.structures.inputs):
            print('Extracting Global PDB Metrics: {}'.format(lab_h))
            # Extract global statistics
            r_info = pdb_i.get_r_rfree_sigma()
            # Populate tables
            self.tables.structures.set_value(lab_h, 'r-work', r_info.r_work)
            self.tables.structures.set_value(lab_h, 'r-free', r_info.r_free)
            self.tables.structures.set_value(lab_h, 'res-high', r_info.high)
            self.tables.structures.set_value(lab_h, 'res-low',  r_info.low)

    def calculate_structure_mean_b_factors(self):
        """Calculate global B-factors for a structure"""
        # ----------------------------------------------------->
        self.tables.structures['mean-b-all']       = numpy.nan
        self.tables.structures['mean-b-protein']   = numpy.nan
        self.tables.structures['mean-b-backbone']  = numpy.nan
        self.tables.structures['mean-b-sidechain'] = numpy.nan
        # ----------------------------------------------------->
        self.tables.structures['rmsd-b-all']       = numpy.nan
        self.tables.structures['rmsd-b-protein']   = numpy.nan
        self.tables.structures['rmsd-b-backbone']  = numpy.nan
        self.tables.structures['rmsd-b-sidechain'] = numpy.nan
        # ----------------------------------------------------->
        print('------------------------------------>')
        for lab_h, pdb_h in zip(self.structures.labels, self.structures.hierarchies):
            print('Calculating Global Mean B-Factors: {}'.format(lab_h))
            # Extract B-factor statistics
            b_stats = BfactorStatistics.from_pdb(pdb_hierarchy=pdb_h.deep_copy())
            # Populate tables
            self.tables.structures.set_value(lab_h, 'mean-b-all',       b_stats.all.mean        )
            self.tables.structures.set_value(lab_h, 'mean-b-protein',   b_stats.protein.mean    )
            self.tables.structures.set_value(lab_h, 'mean-b-backbone',  b_stats.backbone.mean   )
            self.tables.structures.set_value(lab_h, 'mean-b-sidechain', b_stats.sidechain.mean  )
            # ----------------------------------------------------->
            self.tables.structures.set_value(lab_h, 'rmsd-b-all',       b_stats.all.biased_standard_deviation        )
            self.tables.structures.set_value(lab_h, 'rmsd-b-protein',   b_stats.protein.biased_standard_deviation    )
            self.tables.structures.set_value(lab_h, 'rmsd-b-backbone',  b_stats.backbone.biased_standard_deviation   )
            self.tables.structures.set_value(lab_h, 'rmsd-b-sidechain', b_stats.sidechain.biased_standard_deviation  )

    ######################################################################################

    def extract_residue_info(self):
        """Extract information about each of the residues"""
        # ----------------------------------------------------->
        self.tables.residue_observations.loc[:,:,'num_conformers'] = numpy.nan
        self.tables.residue_observations.loc[:,:,'num_atoms']      = numpy.nan
        self.tables.residue_observations.loc[:,:,'occupancy'] = numpy.nan
        # ----------------------------------------------------->
        print('------------------------------------>')
        for lab_h, pdb_h in zip(self.structures.labels, self.structures.hierarchies):
            print('Extracting Residue Info: {}'.format(lab_h))
            for c in conformers_via_residue_groups(pdb_h):
                res_lab = make_label(c)
                # Extract counts for residue group
                self.tables.residue_observations.set_value(res_lab, lab_h, 'num_conformers', len(c.parent().conformers()))
                self.tables.residue_observations.set_value(res_lab, lab_h, 'num_atoms', c.atoms_size())
                # Extract occupancy
                p_occ = [o for o in c.atoms().extract_occ() if o<1.0]
                if c.altloc and (len(set(p_occ)) == 1):
                    self.tables.residue_observations.set_value(res_lab, lab_h, 'occupancy', p_occ[0])

    def calculate_residue_mean_b_factors(self):
        """Extract Mean-B values in each of the structures"""
        # ----------------------------------------------------->
        self.tables.residue_observations.loc[:,:,'mean-b-all']       = numpy.nan
        self.tables.residue_observations.loc[:,:,'mean-b-backbone']  = numpy.nan
        self.tables.residue_observations.loc[:,:,'mean-b-sidechain'] = numpy.nan
        # ----------------------------------------------------->
        print('------------------------------------>')
        for lab_h, pdb_h in zip(self.structures.labels, self.structures.hierarchies):
            print('Calculating Local Mean B-Factors: {}'.format(lab_h))
            cache = pdb_h.atom_selection_cache()
            # Non-Hydrogens
            for c in conformers_via_residue_groups(s_select.non_h(hierarchy=pdb_h, cache=cache)):
                res_lab = make_label(c)
                res_mean_b = flex.mean_weighted(c.atoms().extract_b(), c.atoms().extract_occ())
                self.tables.residue_observations.set_value(res_lab, lab_h, 'mean-b-all', res_mean_b)
            # Backbone Atoms
            for c in conformers_via_residue_groups(s_select.backbone(hierarchy=pdb_h, cache=cache)):
                res_lab = make_label(c)
                res_mean_b = flex.mean_weighted(c.atoms().extract_b(), c.atoms().extract_occ())
                self.tables.residue_observations.set_value(res_lab, lab_h, 'mean-b-backbone', res_mean_b)
            # Sidechain Atoms
            for c in conformers_via_residue_groups(s_select.sidechains(hierarchy=pdb_h, cache=cache)):
                res_lab = make_label(c)
                res_mean_b = flex.mean_weighted(c.atoms().extract_b(), c.atoms().extract_occ())
                self.tables.residue_observations.set_value(res_lab, lab_h, 'mean-b-sidechain', res_mean_b)

    def calculate_residue_mean_normalised_b_factors(self):
        """Extract Mean-B values in each of the structures"""
        # ----------------------------------------------------->
        self.tables.residue_observations.loc[:,:,'mean-bz-all']       = numpy.nan
        self.tables.residue_observations.loc[:,:,'mean-bz-backbone']  = numpy.nan
        self.tables.residue_observations.loc[:,:,'mean-bz-sidechain'] = numpy.nan
        # ----------------------------------------------------->
        print('------------------------------------>')
        for lab_h, pdb_h in zip(self.structures.labels, self.structures.hierarchies):
            print('Calculating Local Normalised Mean B-Factors: {}'.format(lab_h))
            # Normalise the b-factors of the structure
            pdb_h_z = normalise_b_factors_to_z_scores(pdb_hierarchy=pdb_h, method='protein')
            cache = pdb_h_z.atom_selection_cache()
            # Non-Hydrogens
            for c in conformers_via_residue_groups(s_select.non_h(hierarchy=pdb_h_z, cache=cache)):
                res_lab = make_label(c)
                res_mean_b = flex.mean_weighted(c.atoms().extract_b(), c.atoms().extract_occ())
                self.tables.residue_observations.set_value(res_lab, lab_h, 'mean-bz-all', res_mean_b)
            # Backbone Atoms
            for c in conformers_via_residue_groups(s_select.backbone(hierarchy=pdb_h_z, cache=cache)):
                res_lab = make_label(c)
                res_mean_b = flex.mean_weighted(c.atoms().extract_b(), c.atoms().extract_occ())
                self.tables.residue_observations.set_value(res_lab, lab_h, 'mean-bz-backbone', res_mean_b)
            # Sidechain Atoms
            for c in conformers_via_residue_groups(s_select.sidechains(hierarchy=pdb_h_z, cache=cache)):
                res_lab = make_label(c)
                res_mean_b = flex.mean_weighted(c.atoms().extract_b(), c.atoms().extract_occ())
                self.tables.residue_observations.set_value(res_lab, lab_h, 'mean-bz-sidechain', res_mean_b)

    def calculate_normalised_normalised_residue_b_factors(self):
        """Calculate twice-normalised residue b-factors"""

        pass

    def reduce_residue_bfactor_observations(self):
        """Calculate statistics of the bfactor variation"""
        # ----------------------------------------------------->
        self.tables.residue_statistics['mean-mean-bz-all'] = numpy.nan
        self.tables.residue_statistics['rmsd-mean-bz-all'] = numpy.nan
        # ----------------------------------------------------->
        self.tables.residue_statistics['mean-mean-bz-backbone'] = numpy.nan
        self.tables.residue_statistics['rmsd-mean-bz-backbone'] = numpy.nan
        # ----------------------------------------------------->
        self.tables.residue_statistics['mean-mean-bz-sidechain'] = numpy.nan
        self.tables.residue_statistics['rmsd-mean-bz-sidechain'] = numpy.nan
        # ----------------------------------------------------->
        print('------------------------------------>')
        for res_lab in self.tables.residue_observations.items:
            print('Calculating Mean and RMS B-factors: {}'.format(res_lab))
            # Extract the residue data
            res_data = self.tables.residue_observations[res_lab]
            res_mean_bz_all  = res_data['mean-bz-all']
            res_mean_bz_back = res_data['mean-bz-backbone']
            res_mean_bz_side = res_data['mean-bz-sidechain']
            # All atoms
            if res_mean_bz_all.any():
                self.tables.residue_statistics.set_value(res_lab, 'mean-mean-bz-all', numpy.mean(res_mean_bz_all))
                self.tables.residue_statistics.set_value(res_lab, 'rmsd-mean-bz-all', numpy.std(res_mean_bz_all))
            # Backbone atoms
            if res_mean_bz_back.any():
                self.tables.residue_statistics.set_value(res_lab, 'mean-mean-bz-backbone', numpy.mean(res_mean_bz_back))
                self.tables.residue_statistics.set_value(res_lab, 'rmsd-mean-bz-backbone', numpy.std(res_mean_bz_back))
            # Sidechain atoms
            if res_mean_bz_side.any():
                self.tables.residue_statistics.set_value(res_lab, 'mean-mean-bz-sidechain', numpy.mean(res_mean_bz_side))
                self.tables.residue_statistics.set_value(res_lab, 'rmsd-mean-bz-sidechain', numpy.std(res_mean_bz_side))

    def calculate_residue_phi_psi_angles(self):
        """Extract phi-psi angles for each of the structures"""
        if not self.structures.hierarchies: raise Exception('No input objects given')
        # ----------------------------------------------------->
        self.tables.residue_observations.loc[:,:,'phi'] = numpy.nan
        self.tables.residue_observations.loc[:,:,'psi'] = numpy.nan
        # ----------------------------------------------------->
        print('------------------------------------>')
        for lab_h, pdb_h in zip(self.structures.labels, self.structures.hierarchies):
            print('Extracting Phi-Psi Angles: {}'.format(lab_h))
            # Note these are residue objects so model->chain->conformer->residue
            for res, (phi, psi) in get_all_phi_psi_for_hierarchy(pdb_h=pdb_h):
                # Create residue label - need to account for fact that all single confs are duplicated
                if res.is_pure_main_conf: res_lab = (res.parent().parent().id, res.resid(), '')
                else:                     res_lab = (res.parent().parent().id, res.resid(), res.parent().altloc)
                # Check to see if already done so only populate once
                if numpy.isnan(self.tables.residue_observations.get_value(res_lab, lab_h, 'phi')):
                    self.tables.residue_observations.set_value(res_lab, lab_h, 'phi', phi)
                # Populate values
                if numpy.isnan(self.tables.residue_observations.get_value(res_lab, lab_h, 'psi')):
                    self.tables.residue_observations.set_value(res_lab, lab_h, 'psi', psi)

    def reduce_residue_phi_psi_observations(self):
        """Calculate statistics of the phi-psi variation"""
        # ----------------------------------------------------->
        self.tables.residue_statistics['mean-phi'] = numpy.nan
        self.tables.residue_statistics['mean-psi'] = numpy.nan
        self.tables.residue_statistics['rmsd-phi'] = numpy.nan
        self.tables.residue_statistics['rmsd-psi'] = numpy.nan
        self.tables.residue_statistics['rms-rmsd-phi-rmsd-psi'] = numpy.nan
        self.tables.residue_statistics['max-rmsd-phi-rmsd-psi'] = numpy.nan
        # ----------------------------------------------------->
        print('------------------------------------>')
        for res_lab in self.tables.residue_observations.items:
            # Extract the residue data
            res_data = self.tables.residue_observations[res_lab]
            res_phi  = res_data['phi']
            res_psi  = res_data['psi']
            if (not res_phi.any()) and (not res_phi.any()): continue
            print('Calculating RMS Phi-Psi Angles: {}'.format(res_lab))
            # Analyse phi angles
            phi_m, phi_s = mean_std_angles(angles=res_phi)
            self.tables.residue_statistics.set_value(res_lab, 'mean-phi', phi_m)
            self.tables.residue_statistics.set_value(res_lab, 'rmsd-phi', phi_s)
            # Analyse psi angles
            psi_m, psi_s = mean_std_angles(angles=res_psi)
            self.tables.residue_statistics.set_value(res_lab, 'mean-psi', psi_m)
            self.tables.residue_statistics.set_value(res_lab, 'rmsd-psi', psi_s)
            # Combine into one value
            self.tables.residue_statistics.set_value(res_lab, 'rms-rmsd-phi-rmsd-psi', numpy.sqrt(psi_s**2 + phi_s**2))
            self.tables.residue_statistics.set_value(res_lab, 'max-rmsd-phi-rmsd-psi', numpy.max([psi_s, phi_s]))




import os, sys, copy, re, shutil

import libtbx.phil
from bamboo.common.command import CommandManager

#######################################

blank_arg_prepend = {'.mtz':'input.mtz=', None:'column.label='}

master_phil = libtbx.phil.parse("""
input {
    mtz = None
        .multiple = True
        .type = path
}
options {
    label_suffix = *incremental filename foldername
        .help = 'suffix added to the column name in the output file'
        .type = choice
        .multiple = False
    column {
        label = None
            .help = 'columns to be extracted from the input mtz files'
            .type = str
            .multiple = True
    }
}
output {
    mtz = merged.mtz
        .type = path
    log = merged.log
        .type = path
}
""")

#######################################

def run(params):

    in_mtzs = params.input.mtz
    assert len(params.input.mtz) > 1, 'Need to provide at least one mtz files'
#    assert len(params.input.mtz) < 3, 'Need to provide at most two mtz files'

    out_mtz = params.output.mtz
    out_log = params.output.log

    ###########################################
    # COMMAND LINE COMMANDS
    ###########################################
    cm = CommandManager('cad')

    ###########################################
    # ITERATE THROUGH INPUT MTZS
    ###########################################
    for i_mtz, mtz in enumerate(in_mtzs):
        # Use numbering from 1
        n_mtz = i_mtz + 1
        # Create an mtz suffix
        if   params.options.label_suffix == 'incremental': suffix = '-{}'.format(n_mtz)
        elif params.options.label_suffix == 'filename':    suffix = '-{}'.format(os.path.splitext(os.path.basename(mtz))[0])
        elif params.options.label_suffix == 'foldername':  suffix = '-{}'.format(os.path.basename(os.path.dirname(mtz)))
        else:                                              raise Exception('Not yet implemented, sorry')
        # Add to the command line
        cm.add_command_line_arguments(['hklin{}'.format(n_mtz), mtz])
        # Select column labels
        cm.add_standard_input('labin file_number {0} {1}'.format(n_mtz, ' '.join(['E{0}={1}'.format(i+1,c) for i,c in enumerate(params.options.column.label)])))
        cm.add_standard_input('labout file_number {0} {1}'.format(n_mtz, ' '.join(['E{0}={1}'.format(i+1,c+suffix) for i,c in enumerate(params.options.column.label)])))

    ###########################################
    # OUTPUT FILES
    ###########################################
    cm.add_command_line_arguments(['hklout', out_mtz])

    ###########################################
    # RUN
    ###########################################
    print 'Running CAD:'
    cm.print_settings()
    ret_code = cm.run()
    cm.write_output(log_file=out_log)
    if ret_code != 0:
        print '============================>'
        print 'Refmac returned with an error'
        print '============================>'
        print cm.output
        print '============================>'
        print cm.error
        print '============================>'

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os, glob

from bamboo.html import png2base64str
from pandda.html import PANDDA_HTML_ENV

def write_initial_html(pandda):
    # Get template to be filled in
    template = PANDDA_HTML_ENV.get_template('pandda_summary.html')
    # Output file
    out_file = pandda.output_handler.get_file(file_tag='initial_html')
    # Output directory (for relative symlinks)
    out_dir  = os.path.abspath(os.path.dirname(out_file))

    # ===========================================================>
    # Construct the data object to populate the template
    output_data = {}
    output_data['header'] = 'PANDDA Dataset Summaries'
    output_data['title'] = 'PANDDA Dataset Summaries'
    output_data['introduction'] = 'Summary of Added Datasets'
    # ===========================================================>
    # Summary bar
    output_data['summary_bar'] = []
    output_data['summary_bar'].append({'colour':'info',    'text':'Datasets Loaded:   {}'.format(pandda.datasets.size())                                            })
    output_data['summary_bar'].append({'colour':'success', 'text':'Datasets Accepted: {}'.format(pandda.datasets.size(mask_name='rejected - total', invert=True))   })
    output_data['summary_bar'].append({'colour':'danger',  'text':'Datasets Rejected: {}'.format(pandda.datasets.size(mask_name='rejected - total'))                })
    # ===========================================================>
    # Header Images
    output_data['top_images'] = []
    output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=pandda.output_handler.get_file(file_tag='d_resolutions'))),
                                       'title': 'Dataset Resolutions' })
    output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=pandda.output_handler.get_file(file_tag='d_rfactors'))),
                                       'title': 'Dataset R-Factors' })
    output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=pandda.output_handler.get_file(file_tag='d_global_rmsd_to_ref'))),
                                       'title': 'Dataset RMSD to Mean Structure' })
    output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=pandda.output_handler.get_file(file_tag='d_cell_volumes'))),
                                       'title': 'Dataset Cell Volumes' })
    output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=pandda.output_handler.get_file(file_tag='d_cell_axes'))),
                                       'title': 'Dataset Cell Axis Lengths' })
    output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=pandda.output_handler.get_file(file_tag='d_cell_angles'))),
                                       'title': 'Dataset Cell Angles' })
    # ===========================================================>
    # Write Output
    with open(out_file, 'w') as out_html:
        out_html.write(template.render(output_data))

def write_analyse_html(pandda):

    # Get template to be filled in
    template = PANDDA_HTML_ENV.get_template('pandda_summary.html')
    # Output file
    out_file = pandda.output_handler.get_file(file_tag='analyse_html')
    # Output directory (for relative symlinks)
    out_dir  = os.path.abspath(os.path.dirname(out_file))

    # Extract the dataset info as a dictionary
    all_d_info = pandda.tables.dataset_info.transpose().to_dict()
    all_m_info = pandda.tables.dataset_map_info.transpose().to_dict()

    len_data = len(pandda.tables.dataset_info.index)
    num_interesting = len([d for d in pandda.datasets.mask(mask_name='rejected - total', invert=True) if d.events])
    num_analysed    = len([d for d in pandda.datasets.mask(mask_name='rejected - total', invert=True) if d.meta.analysed])
    num_rejected    = pandda.datasets.size(mask_name='rejected - total')
    num_not_interesting = num_analysed - num_interesting
    num_not_analysed    = len_data - num_analysed - num_rejected

    # ===========================================================>
    # Construct the data object to populate the template
    output_data = {}
    output_data['header'] = 'PANDDA Processing Output'
    output_data['title'] = 'PANDDA Processing Output'
    output_data['introduction'] = 'Summary of Processing of Datasets'
    # ===========================================================>
    # Summary bar
    output_data['summary_bar'] = []
    output_data['summary_bar'].append({'colour':'info',    'text':'Analysed: {}'.format(num_analysed)               })
    output_data['summary_bar'].append({'colour':'success', 'text':'Interesting: {}'.format(num_interesting)         })
    output_data['summary_bar'].append({'colour':'warning', 'text':'Not Interesting: {}'.format(num_not_interesting) })
    output_data['summary_bar'].append({'colour':'danger',  'text':'Not Analysed: {}'.format(num_not_analysed)       })
    # Progress Bars
    output_data['progress_bar'] = []
    output_data['progress_bar'].append({'title':'Dataset Summary', 'data':[]})
    output_data['progress_bar'][0]['data'].append({'text':'Interesting',     'colour':'success', 'size':100.0*num_interesting/len_data})
    output_data['progress_bar'][0]['data'].append({'text':'Not Interesting', 'colour':'info',    'size':100.0*num_not_interesting/len_data})
    output_data['progress_bar'][0]['data'].append({'text':'Not Analysed',    'colour':'warning', 'size':100.0*num_not_analysed/len_data})
    output_data['progress_bar'][0]['data'].append({'text':'Rejected',        'colour':'danger',  'size':100.0*num_rejected/len_data})
    # ===========================================================>
    # Header Images
    output_data['top_images'] = []
    if os.path.exists(pandda.output_handler.get_file(file_tag='pymol_sites_png_1')):
        output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=pandda.output_handler.get_file(file_tag='pymol_sites_png_1'))),
                                           'title': 'Identified Sites (Front)' })
    if os.path.exists(pandda.output_handler.get_file(file_tag='pymol_sites_png_2')):
        output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=pandda.output_handler.get_file(file_tag='pymol_sites_png_2'))),
                                           'title': 'Identified Sites (Back)' })
    for i_png, png in enumerate(sorted(glob.glob(pandda.output_handler.get_file(file_tag='analyse_site_graph_mult').format('*')))):
        output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=png)),
                                           'title': 'Identified Site Summary ({})'.format(i_png+1) })
    # ===========================================================>
    # Tables
    output_data['table'] = {}
    output_data['table']['column_headings'] = ['Dataset', 'Crystal', 'Structure', 'RFree/RWork', 'Resolution', 'Reference RMSD', 'Map Uncertainty', 'Overall', 'Interesting Areas', 'Result']
    output_data['table']['rows'] = []
    # Add the datasets as rows
    for d in pandda.datasets.all():

        d_data = all_d_info[d.tag]
        m_data = all_m_info[d.tag]
        rmsd  = d_data['rmsd_to_reference']
        rfree = d_data['r_free']
        rwork = d_data['r_work']
        res_h = d_data['high_resolution']
        uncty = m_data['map_uncertainty']

        rejection_reason = d_data['rejection_reason']

        # colour choices - 'success', 'muted', 'danger'
        # icon choices   - 'ok', 'flag', 'remove'

        columns = []
        # ------------------------------>>>
        # Test for Bad Crystal
        # ------------------------------>>>
        if pandda.datasets.all_masks().get_mask_value(mask_name='rejected - crystal', entry_id=d.tag) == True:
            columns.append({'colour':'danger',  'icon':'remove',    'message':rejection_reason})
        else:
            columns.append({'colour':'success', 'icon':'ok',        'message':'OK'})
        # ------------------------------>>>
        # Test for Bad Structures
        # ------------------------------>>>
        if pandda.datasets.all_masks().get_mask_value(mask_name='rejected - structure', entry_id=d.tag) == True:
            columns.append({'colour':'danger',  'icon':'remove',    'message':rejection_reason})
        else:
            columns.append({'colour':'success', 'icon':'ok',        'message':'OK'})
        # ------------------------------>>>
        # Test for Refinement Success - some test on r-free
        # ------------------------------>>>
        if pandda.datasets.all_masks().get_mask_value(mask_name='bad crystal - rfree', entry_id=d.tag) == True:
            columns.append({'colour':'danger',  'icon':'remove',    'message':'{}/{}'.format(rfree,rwork)})
        else:
            columns.append({'colour':'success', 'icon':'default',   'message':'{}/{}'.format(rfree,rwork)})
        # ------------------------------>>>
        # Resolution
        # ------------------------------>>>
        if 0:
            columns.append({'colour':'danger',  'icon':'remove',    'message':'{}'.format(res_h)})
        else:
            columns.append({'colour':'success', 'icon':'ok',        'message':'{}'.format(res_h)})
        # ------------------------------>>>
        # Test for Structure movement
        # ------------------------------>>>
        if pandda.datasets.all_masks().get_mask_value(mask_name='bad crystal - isomorphous structure', entry_id=d.tag) == True:
            columns.append({'colour':'danger',  'icon':'remove',    'message':'{}'.format(rmsd)})
        else:
            columns.append({'colour':'success', 'icon':'ok',        'message':'{}'.format(rmsd)})
        # ------------------------------>>>
        # Uncertainty of Map
        # ------------------------------>>>
        if 0:
            columns.append({'colour':'danger',  'icon':'remove',    'message':'{}'.format(uncty)})
        else:
            columns.append({'colour':'success', 'icon':'ok',        'message':'{}'.format(uncty)})
        # ------------------------------>>>
        # Analysed
        # ------------------------------>>>
        if d.meta.analysed == True:
            columns.append({'colour':'success', 'icon':'ok',        'message':'Analysed'})
        else:
            columns.append({'colour':'danger',  'icon':'remove',    'message':'Not Analysed'})
        # ------------------------------>>>
        # Test for if it's interesting
        # ------------------------------>>>
        if d.events:
            columns.append({'colour':'success', 'icon':'ok',        'message':'{} Clusters'.format(len(d.events))})
        else:
            columns.append({'colour':'default', 'icon':'remove',    'message':''})

        row_message = 'Z-Blobs Found' if d.events else \
                      ''
        row_colour  = 'success' if row_message else \
                      'danger' if pandda.datasets.all_masks().get_mask_value(mask_name='rejected - total', entry_id=d.tag) == True else \
                      'default'

        output_data['table']['rows'].append({'heading' : d.tag,
                                             'colour'  : row_colour,
                                             'message' : row_message,
                                             'columns' : columns})

    # ===========================================================>
    # Write Output
    with open(out_file, 'w') as out_html:
        out_html.write(template.render(output_data))



try:
    from pymol import cmd
    from pymol.cgo import *
except:
    pass

class Shape(object):
    def draw(self, name):
        print 'Loading %s to %s ' % (self.SHAPE, name)
        exec(self.as_cmd(name))
    def as_cmd(self, name):
        return 'cmd.load_cgo(%s, "%s")' % (self.as_string(), name)
    def as_string(self):
        raise Exception('Not Defined')

class Cylinder(Shape):
    SHAPE = 'Cylinder'
    def __init__(self, start, end, colors=None, radius=None):
        if (colors is None):    colors = [[1,0,0]]*2
        if (radius is None):     radius  = 0.1
        self.start = start
        self.end = end
        self.colors = colors
        self.radius = radius

    def as_string(self):
        s = "["
        s += "CYLINDER," + " %.6g, %.6g, %.6g," % tuple(self.start)
        s += " %.6g, %.6g, %.6g," % tuple(self.end)
        s += " %.6g," % self.radius
        s += " %.6g, %.6g, %.6g," % tuple(self.colors[0])
        s += " %.6g, %.6g, %.6g" % tuple(self.colors[1])
        s += "]"
        return s

class Sphere(Shape):
    SHAPE = 'Sphere'
    def __init__(self, centre, radius=None):
        if (radius is None):    radius  = 1
        self.centre = centre
        self.radius = radius
#        self.color = color
    def as_string(self):
        s = "["
        s += "SPHERE,"
        s += " %.6g, %.6g, %.6g," % tuple(self.centre)
        s += " %.6g" % self.radius
        s += "]"
        return s




import os, sys

from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.QtWebKit import *

class panddaHtmlWidget(QWidget):
    def __init__(self, name, content):
        super(panddaHtmlWidget, self).__init__()
        # Widget Name
        self.name = name
        self.content = content

        # Add a VBox to the panddaWidget
        self.layout = QVBoxLayout(self)
        self.view = None

#        # Add refresh button
#        self.refresh_button = QPushButton('Refresh', self)
#        self.refresh_button.clicked.connect(self.refresh)
#        # Add the button to the vbox
#        self.layout.addWidget(self.refresh_button)

        # Add the html
        self.build_html()

    def build_html(self):
        # Delete old view
        if self.view: self.view.deleteLater()
        # Add a view to the panddaWidget
        self.view = QWebView(self)
        self.layout.addWidget(self.view)
        # Add html to the view
        if os.path.exists(self.content):
            with open(self.content, 'r') as fh:
                text = fh.read()
            dir = os.path.dirname(os.path.abspath(self.content))
            # Dynamically change the html paths to full paths to allow rendering
            text = text.replace('src="../', 'src="file://'+dir+'/../')
            text = text.replace('src="./', 'src="file://'+dir+'/')
            # Set the text in the html window
            self.view.setHtml("<html><head></head><body></body></html>")
            self.view.setHtml(text)
        else:
            self.view.setHtml('File does not yet exist: {}'.format(self.content))
        self.view.show()

    def refresh(self):
        print 'Refreshing html'
        self.build_html()

class panddaWindow(QTabWidget):
    def __init__(self, tabs=None):
        super(panddaWindow, self).__init__()
        self.setWindowTitle('PANDDA GUI')
        # Create Tabs
        self.tabs = []
        if tabs is not None:
            for t in tabs:
                self.add_tab(t)
    def add_tab(self, widget):
        self.tabs.append(widget)
        self.addTab(widget, widget.name)

#class Render(QWebPage):
#  def __init__(self, url):
#    self.app = QApplication(sys.argv)
#    QWebPage.__init__(self)
#    self.loadFinished.connect(self._loadFinished)
#    self.mainFrame().load(QUrl(url))
#    self.app.exec_()
#
#  def _loadFinished(self, result):
#    self.frame = self.mainFrame()
#    self.app.quit()

def test_1(test_dir):
    app = QApplication(sys.argv)
    window = panddaWindow()
    window.add_tab(panddaHtmlWidget(name='RESULTS', content=os.path.join(test_dir,'./pandda_results_example_basic.html')))
    window.show()
    app.exec_()

#def test_2(test_dir):
#    url = path2url(os.path.join(test_dir,'pandda_results_example.html'))
#    print url
#    html = Render(url).frame.toHtml()
#    print html
#
#    app = QApplication(sys.argv)
#    window = panddaWindow()
#    widget = panddaHtmlWidget(name='RESULTS', content=html)
#    window.add_tab(widget)
#    window.show()
#    app.exec_()

if __name__ == "__main__":
    test_dir = os.path.join(os.path.dirname(__file__), 'test_files')
    test_1(test_dir)
    test_2(test_dir)



"""Contains strings used to build up a coot script"""

COOT_exit = """coot_real_exit(1)"""

COOT_load_pdb = """read_pdb('<pdbin>')"""
COOT_load_mtz = """auto_read_make_and_draw_maps('<mtzin>')"""

COOT_real_space_refine = """regularize_zone(<imol>, <chain_id>, <resno1>, <resno2>, <altconf>)"""



import os, sys, glob, time, re

from scipy import spatial
import numpy

from libtbx import easy_mp
from libtbx.math_utils import ifloor, iceil
from scitbx.array_family import flex
import iotbx.crystal_symmetry_from_any

from cctbx import crystal

from giant.structure import make_label
from giant.grid.utils import create_cartesian_grid

class grid_handler(object):
    def __init__(self, verbose=True):
        """Create and manage a grid object to be sampled across many aligned datasets"""

        self.verbose = verbose

        # Size of complete grid
        self._grid_size = None
        self._grid_spacing = None

        # Cartesian values of grid
        self._cart_max = None
        self._cart_min = None
        self._cart_size = None

        # Atomic Masks for partitioning the grid
        self._grid_masks = {}
        # Local mask for filtering/smoothing
        self._local_mask = None

        # Partiton object for spliting the grid into sections around atoms
        self._grid_partition = None

    def set_grid_spacing(self, spacing):
        self._grid_spacing = spacing
    def grid_spacing(self):
        return self._grid_spacing
    def grid_size(self):
        return self._grid_size
    def grid_size_1d(self):
        return self._grid_size[0]*self._grid_size[1]*self._grid_size[2]
    def grid_point_volume(self):
        return self.grid_spacing()**3

    def set_cart_extent(self, cart_min, cart_max):
        self._cart_min = cart_min
        self._cart_max = cart_max
        self._cart_size = tuple([s1 - s2 for s1,s2 in zip(cart_max, cart_min)])
    def cart_max(self):
        return self._cart_max
    def cart_min(self):
        return self._cart_min
    def cart_extent(self):
        return (self._cart_min, self._cart_max)
    def cart_size(self):
        return self._cart_size

    def cart_points(self):
        return flex.vec3_double(flex.nested_loop(grid_size)) * self.grid_spacing() + self.cart_min()
    def grid_points(self):
        return flex.nested_loop(self.grid_size())

    def grid_indexer(self):
        """Translates between 3d grid coordinates and 1d array coordinates"""
        return flex.grid(self.grid_size())

    def unit_cell(self):
        """Create a unit cell as if the reference grid were a real lattice (THE UNIT CELL IS LARGER THAN cart_size() BY 1 GRID SPACING)"""
        return crystal.uctbx.unit_cell('{!s} {!s} {!s} 90 90 90'.format(*list(flex.double(self.grid_size())*self.grid_spacing())))
    def space_group(self):
        """Create a spacegroup as if the reference grid were a real lattice (Cartesian Grid == P1 SpaceGroup)"""
        return crystal.sgtbx.space_group('P1')
    def symmetry(self):
        """Create a symmetry object as if the reference grid were a real lattice (THE UNIT CELL IS LARGER THAN cart_size() BY 1 GRID SPACING)"""
        return iotbx.crystal_symmetry_from_any.from_string("{:f},{:f},{:f},90,90,90,P1".format(*list(flex.double(self.grid_size())*self.grid_spacing())))

    def create_grid_partition(self, atomic_hierarchy):
        """Partition the grid using nearest neighbour algorithm"""
        self._grid_partition = grid_partition(  grid_size = self.grid_size(),
                                                grid_spacing = self.grid_spacing(),
                                                atomic_hierarchy = atomic_hierarchy  )
    def partition(self):
        return self._grid_partition

    def add_mask(self, mask_name, mask):
        """Add a an atomic mask to the reference grid"""
        assert mask_name not in self._grid_masks.keys(), 'MASK ALREADY ADDED: {!s}'.format(mask_name)
        self._grid_masks[mask_name] = mask
    def mask(self, mask_name):
        """Return a named atomic mask"""
        return self._grid_masks[mask_name]

    def set_global_mask(self, mask):
        """Add a global mask to the grid object - This will create binary mask of the masked grid points"""
        self._grid_masks['protein'] = mask
        print self.global_mask().summary()
    def global_mask(self):
        return self._grid_masks.get('protein', None)

    def set_symmetry_mask(self, mask):
        """Add a global mask to the grid object - This will create binary mask of the masked grid points"""
        self._grid_masks['symmetry'] = mask
        print self.symmetry_mask().summary()
    def symmetry_mask(self):
        return self._grid_masks.get('symmetry', None)

    def set_local_mask(self, mask):
        """Add local mask to the grid object"""

        self._local_mask = mask
        print self.local_mask().summary()

#        grid_size = self.grid_size()
#        grid_jump = self.local_mask().grid_jump()
#        grid_buffer = self.local_mask().buffer_size()
#        grid_indexer = self.grid_indexer()
#
#        # Resample grid points
#        self._resampled_grid_points = [gp for gp in flex.nested_loop(grid_size) if not [1 for coord in gp if (coord%grid_jump != 0)]]
#        self._resampled_grid_size = tuple([int(1+(g-1)/grid_jump) for g in grid_size])
#        self._resampled_grid_spacing = grid_jump*self.grid_spacing()
#
#        # Create a buffer zone at the edge of the grid
#        buffer_mask_points = [gp for gp in flex.nested_loop(grid_size) if [1 for i_dim, coord in enumerate(gp) if (coord<grid_buffer or coord>(grid_size[i_dim]-1-grid_buffer))]]
#
#        # Create binary mask for the buffer zone
#        buffer_mask_indices = []
#        [buffer_mask_indices.append(grid_indexer(gp)) for gp in buffer_mask_points]
#
#        self._buffer_mask_indices = numpy.array(buffer_mask_indices)

    def local_mask(self):
        return self._local_mask

    def summary(self):
        return '\n'.join([  '----------------------------------->>>',
                            'Reference Grid Summary:',
                            'Grid Spacing:        {!s}'.format(round(self.grid_spacing(), 3)),
                            'Grid Point Volume:   {!s}'.format(round(self.grid_point_volume(),3)),
                            'Size of Grid (3D):   {!s}'.format(self.grid_size()),
                            'Size of Grid (1D):   {!s}'.format(self.grid_size_1d()),
                            'Min of Grid (Cart): {!s}'.format(tuple([round(x,3) for x in self.cart_min()])),
                            'Max of Grid (Cart): {!s}'.format(tuple([round(x,3) for x in self.cart_max()])),
                            'Size of Grid (Cart): {!s}'.format(tuple([round(x,3) for x in self.cart_size()]))
                        ])

    def create_cartesian_grid(self, expand_to_origin):
        if expand_to_origin:
            # TODO Don't know if this will still work - raise error for now
            raise Exception('NOT CURRENTLY CHECKED')
            assert [i>0 for i in self.cart_min()], 'ALL GRID SITES MUST BE GREATER THAN 0 IF ORIGIN INCLUDED'
            assert [i>0 for i in self.cart_max()], 'ALL GRID SITES MUST BE GREATER THAN 0 IF ORIGIN INCLUDED'
            box_size, self._grid_size, cart_points = create_cartesian_grid(min_carts=(0,0,0),
                                                                                 max_carts=self.cart_max(),
                                                                                 grid_spacing=self.grid_spacing())
        else:
            box_size, self._grid_size, cart_points = create_cartesian_grid(min_carts=self.cart_min(),
                                                                                 max_carts=self.cart_max(),
                                                                                 grid_spacing=self.grid_spacing())

        # Update max/min cart sizes as the grid will be slightly larger than the requested size
        self.set_cart_extent(cart_min=cart_points.min(), cart_max=cart_points.max())

        return self.grid_size()

class grid_partition(object):
    def __init__(self, grid_size, grid_spacing, atomic_hierarchy):
        """Partition a grid based on the nearest neighbour calpha atom for each grid site"""

        # Save inputs
        self.hierarchy = atomic_hierarchy
        self.grid_size = grid_size
        self.grid_spacing = grid_spacing
        self.grid_indexer = flex.grid(grid_size)

        # Calculate partition variables
        atoms = [at for at in atomic_hierarchy.atoms_with_labels()]
        #self.grid_sites = list(flex.nested_loop(grid_size))
        self.atom_sites_grid = numpy.array([a.xyz for a in atoms])/grid_spacing
        self.atom_sites_cart = numpy.array([a.xyz for a in atoms])
        self.atom_labels = [make_label(a) for a in atoms]

        self._atoms = atomic_hierarchy.deep_copy()

        # Distances from grid sites to nearest atom site
        self.nn_dists = None
        # Index of nearest atom to grid sites
        self.nn_groups = None
        # Label of nearest atom to grid sites
        #self.nn_atom_labels = None

    def grid_sites(self):
        return list(flex.nested_loop(self.grid_size))

    def partition_grid(self, cpus=1):
        """Find the nearest neighbour for each grid point"""

#        tree = spatial.KDTree(data=self.atom_sites_grid)
#        self.nn_dists, self.nn_groups = tree.query(self.grid_sites())
#        self.nn_atom_labels = [self.atom_labels[i] for i in self.nn_groups]

        def find_sites(sites_dict):
            ref_sites   = sites_dict['ref']
            query_sites = sites_dict['query']
            tree = spatial.KDTree(data=ref_sites)
            nn_dists, nn_groups = tree.query(query_sites)
            return {'nn_dists':nn_dists, 'nn_groups':nn_groups}

        assert isinstance(cpus, int)
        assert cpus > 0

        # Points that define neighbourhoods
        ref_sites = self.atom_sites_grid
        # Sites that we are partitioning
        query_sites = self.grid_sites()

        if cpus == 1:
            output = [find_sites({'ref':ref_sites, 'query':query_sites})]
        else:
            # Chunk the points into groups
            chunk_size = iceil(1.0*len(query_sites)/cpus)
            chunked_points = [query_sites[i:i + chunk_size] for i in range(0, len(query_sites), chunk_size)]
            assert sum([len(a) for a in chunked_points]) == len(query_sites)
            assert len(chunked_points) == cpus

            # Map to cpus
            arg_list = [{'ref':ref_sites, 'query':chunk} for chunk in chunked_points]
            output = easy_mp.pool_map(fixed_func=find_sites, args=arg_list, processes=cpus)

        assert len(output) == cpus, '{!s} != {!s}'.format(len(output), cpus)

        self.nn_dists = [];  [self.nn_dists.extend(t['nn_dists']) for t in output]
        self.nn_dists = numpy.array(self.nn_dists)
        assert len(query_sites) == len(self.nn_dists)
        self.nn_groups = []; [self.nn_groups.extend(t['nn_groups']) for t in output]
        self.nn_groups = numpy.array(self.nn_groups)
        assert len(query_sites) == len(self.nn_groups)
        #self.nn_atom_labels = [self.atom_labels[i] for i in self.nn_groups]

    def query_by_grid_indices(self, idxs):
        """Return the atom label for a grid site index"""
        if self.nn_groups is None: self.partition_grid()
        return [self.atom_labels[self.nn_groups[i]] for i in idxs]

    def query_by_grid_points(self, gps):
        """Return the atom label for a grid point"""
        if self.nn_groups is None: self.partition_grid()
        return [self.atom_labels[self.nn_groups[self.grid_indexer(g)]] for g in gps]

    def query_by_cart_points(self, sites_cart):
        """Dynamically calculate the nearest atom site to the input points"""
        tree = spatial.KDTree(data=self.atom_sites_cart)
        nn_dists, nn_groups = tree.query(sites_cart)
        return [self.atom_labels[i] for i in nn_groups]




from bamboo.common import Meta, Info

########################################################################################################
#
#   MAP LIST CLASSES
#
########################################################################################################

class MapList(Info):
    _map_names = []
    def __init__(self, map_names=None):
        if map_names is None: map_names=[]
        assert self._map_names+map_names, 'No Map Names defined'
        for m in self._map_names+map_names:
            self.__dict__[m] = None
        # Intialise meta dict
        self.meta = Meta()
        # Set initialised so attributes can't be added
        self._initialized = True

class PanddaStatMapList(MapList):
    _map_names = ['mean_map','medn_map','stds_map','sadj_map','skew_map','kurt_map','bimo_map']

class PanddaMultipleStatMapList(object):
    def __init__(self):
        """Store lots of statistical maps"""
        self.map_lists = {}
    def get_resolutions(self):
        return sorted(self.map_lists.keys())
    def add(self, stat_map_list, resolution, overwrite=False):
        assert isinstance(resolution, float), 'Resolution of map must be of type float. Type given: {!s}'.format(type(resolution))
        if not overwrite: assert resolution not in self.map_lists.keys(), 'MAPS OF THIS RESOLUTION ALREADY ADDED'
        assert isinstance(stat_map_list, PanddaStatMapList), 'stat_map_list must be of type PanddaMultipleStatMapList. Type given: {!s}'.format(type(stat_map_list))
        self.map_lists[resolution] = stat_map_list
    def get(self, resolution):
        return self.map_lists[resolution]





class SelectionSummary(object):

    def __init__(self, ref_hierarchy, sub_hierarchy=None, selection=None):
        """Summarise a selection of a hierarchy relative to the whole hierarchy"""

        assert [selection, sub_hierarchy].count(None)==1,'Provide selection OR sub_hierarchy'

        if selection:
            cache = ref_hierarchy.atom_selection_cache()
            lig_sel = cache.selection(ligand_selection)
            sub_hierarchy = hierarchy.select(lig_sel)

        self.hierarchy_ref = ref_hierarchy
        self.hierarchy_sub = sub_hierarchy










import os, sys, copy, glob
import shutil

import libtbx.phil

import numpy

############################################################################

master_phil = libtbx.phil.parse("""
input {
    pandda_dir = './pandda'
        .help = 'Path to the pandda directory to use as a template'
        .type = str
}
output {
    out_dir = './pandda-migrate'
        .help = 'output folder name'
        .type = str
}

verbose = False
    .type = bool

""")

############################################################################

def copy_directory_contents(in_dir, out_dir, regex='*'):

    if isinstance(regex, str):
        regex = [regex]
    else:
        assert isinstance(regex, list), 'regex must either be str or list of str'
        for r in regex: assert isinstance(r, str), 'regex must either be str or list of str'

    if not os.path.exists(out_dir): os.mkdir(out_dir)

    for file_regex in regex:
        for file_in in sorted(glob.glob(os.path.join(in_dir, file_regex))):
            # Create output filename
            file_out = os.path.join(out_dir, os.path.basename(file_in))
            # Copy file
            print 'Copying {} -> {}'.format(file_in, file_out)
            shutil.copy(file_in, file_out)

def migrate_pandda(params):

    # Create output directory
    os.mkdir(params.output.out_dir)

    # Copy across reference structures (copying the file if it's a softlink)
    copy_directory_contents(
        in_dir  = os.path.join(params.input.pandda_dir, 'reference'),
        out_dir = os.path.join(params.output.out_dir,   'reference'),
        regex   = ['*.pdb', '*.mtz', '*.ccp4']
    )

    # Copy across pickles (copying the file if it's a softlink)
    copy_directory_contents(
        in_dir  = os.path.join(params.input.pandda_dir, 'pickled_panddas'),
        out_dir = os.path.join(params.output.out_dir,   'pickled_panddas'),
        regex   = ['reference*.pickle', 'statistical*.pickle']
    )

def run(params):

    assert params.input.pandda_dir, 'Must specify pandda directory'
    assert params.output.out_dir,   'Must specify output directory'
    assert not os.path.exists(params.output.out_dir), 'Output directory already exists'

    print 'Migrating Pandda from {} to {}'.format(params.input.pandda_dir, params.output.out_dir)
    migrate_pandda(params)

#######################################

if __name__ == '__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)




class Meta(object):
    """Object for storing random data - can be edited and added to as required"""
    _initialized = False
    def __init__(self, args=None):
        if isinstance(args, dict):
            for k in args:
                self.__dict__[k] = args[k]
        elif isinstance(args, list):
            for l in args:
                self.__dict__[l] = None
        elif args is not None:
            raise Exception('args must be dict or list')
        _initialized = True

    def summary(self):
        out = []
        for k in self.__dict__:
            out.append('{!s}: {!s}'.format(k, self.__dict__[k]))
        return '\n'.join(out)

class Info(Meta):
    """Same as Meta, but cannot change variables after setting _initialized=True"""
    def __setattr__(self, name, value):
        if self._initialized and (not hasattr(self, name)):
            raise AttributeError('Cannot set new attributes after initialisation: {!s}'.format(name))
        object.__setattr__(self, name, value)

def compare_dictionaries(dict1, dict2):
    """Check differences between dicts. Returns tuple(differences in the keys, and differences in the values)."""

    uniq_keys = list(set(dict1.keys()).symmetric_difference(set(dict2.keys())))
    uniq_summ = [(d_key, (d_key in dict1.keys()), (d_key in dict2.keys())) for d_key in uniq_keys]

    comm_keys = list(set(dict1.keys()).intersection(set(dict2.keys())))
    comm_summ = [(c_key, dict1[c_key], dict2[c_key]) for c_key in comm_keys if dict1[c_key] != dict2[c_key]]

    return uniq_summ, comm_summ

